import datacompy

from data_providers import DataProvider
import logging


class SDTMDataProvider(DataProvider):
    """
    A wrapper-class over DataProvider to extract specifically SDTM data
    """

    RDFSLABEL = "rdfs:label"

    def __init__(self, check_for_refarctored=True, *args, **kwargs):
        self.check_for_refarctored = check_for_refarctored
        super().__init__(*args, **kwargs)

    def get_data_sdtm(self, standard: str, domain: str, study=None, where_map=None, user_role=None):
        assert where_map is None or isinstance(where_map, dict)
        if not where_map:
            where_map = {}
        assert study is None or isinstance(study, str)
        if study:
            where_map = {**where_map, **{
                'Study': {'rdfs:label': study}}}
        # TODO: add assert statements to check prerequisites - e.g. extraction model contains all required nodes and properties
        meta = self.neo_get_meta(standard=standard, table=domain)
        if self.debug:
            print("meta", meta)
        if meta:
            classes = (['Study'] if 'Study' not in meta[0]['classes'] else []) + meta[0]['classes']
            non_valid, no_access = [], []

            if self.check_for_refarctored:
                classes, non_valid = self.neo_validate_classes_to_extract(classes)
                if non_valid:
                    print(
                        f"ERROR: the following classes were excluded as those were never created during refactoring: {non_valid}")
            if user_role:
                classes, no_access = self.neo_validate_access(classes, user_role=user_role)
                print(
                    f"WARNING: the following classes were excluded as the user_role {user_role} access is restricted: {no_access}")
            if meta[0]['req_classes']:
                classes = [(class_ + self.OCLASS_MARKER if class_ not in meta[0]['req_classes'] else class_) for
                           class_ in classes]
            if self.debug:
                print(f"Getting classes: {classes}")

            rels = meta[0]['rels']
            excluded_classes = no_access + non_valid
            if excluded_classes:
                print(f'Excluding the following classes from rels: {excluded_classes}')
                rels = self.filter_classes_from_rels(rels, excluded_classes)

            df = self.get_data_generic(labels=classes,
                                       rels=rels,
                                       where_map=where_map,
                                       infer_rels=False,
                                       return_nodeid=False,
                                       return_propname=False,
                                       use_shortlabel=True,
                                       only_props=[self.RDFSLABEL],
                                       limit=None)

            # renaming and re-ordering columns (according to metadata):
            rename_dct = {}
            for key, item in meta[0]['rename_dct'].items():
                if not item in rename_dct.values():  # to avoid 2 columns with the same name
                    if key != item:
                        rename_dct[key] = item
            df = df.rename(rename_dct, axis=1)

            def _sorter(x):
                # Some columns don't have an order set
                none_count = 1000
                if not x[1]:
                    order = none_count
                    none_count += 1
                else:
                    order = x[1]
                return order

            _col_order = [(k, v) for k, v in sorted(meta[0]['order_dct'].items(), key=lambda x: x[0])]
            col_order = [k for k, v in sorted(_col_order, key=lambda x: _sorter(x))]

            for col in col_order:
                if col not in df.columns:
                    df[col] = None
            df = df[col_order]

            # Sorting
            # check that variables for sorting in meta actaully exist
            sorting, sorting_excluded = [], []
            if meta[0]['sorting']:
                if isinstance(meta[0]['sorting'], list):
                    l_sorting = meta[0]['sorting']
                elif isinstance(meta[0]['sorting'], str):
                    if "," in meta[0]['sorting']:
                        l_sorting = meta[0]['sorting'].split(",")
                    else:
                        l_sorting = [meta[0]['sorting']]
                else:
                    l_sorting = []
                for col in l_sorting:
                    if col in df.columns:
                        sorting.append(col)
                    else:
                        sorting_excluded.append(col)
                    if sorting_excluded:
                        print(
                            f"ERROR: the following columns were excluded from sort-by-group as those have not been extracted"
                            f"from the graph: {sorting_excluded}")

            # sorting dataframe:
            if sorting:
                df = df.sort_values(by=sorting, ignore_index=True)
            else:
                print(f"WARNING: no sort-by-group metadata(`Source Data Table`.SortOrder) was provided")

            return df

    def neo_get_meta(self, standard: str, table: str):

        q = """
        MATCH (sdf:`Data Extraction Standard`{_tag_:$standard})-[:HAS_TABLE]->(sdt:`Source Data Table`{_domain_:$table}),
        (sdt)-[:HAS_COLUMN]->(sdc:`Source Data Column`), (class:Class)<-[:TO]-(rel:Relationship)-[:FROM]->(fromclass:Class)
        WHERE
          EXISTS(
            (sdc)<-[:MAPS_TO_COLUMN]-(rel)-[:TO]->(class)
          )
          OR
          EXISTS(
            (sdt)-[:MAPS_TO_CLASS]->(class)
          )
        WITH *
        OPTIONAL MATCH (sdc)-[r:MAPS_TO_CLASS]->(class)   
        WITH *
        ORDER BY sdf, sdt, sdc.Order, sdc
        RETURN
        collect({from: fromclass.label, to: class.label, type: rel.relationship_type, short_label: class.short_label}) as rels,
        collect(distinct fromclass.label) + collect(distinct class.label) as classes,
        apoc.coll.toSet(
                [triple in [ 
                    triple in collect([class.label, sdc.Core, class.create]) 
                    where triple[1] = "Req" or triple[2]] | triple[0]
                ] + 
                [pair in [
                    pair in collect([fromclass.label, fromclass.create]) 
                    where pair[1]] | pair[0]
                ] // Ensure Domain is in required classes
            ) as req_classes,         
        apoc.map.fromPairs(
            [y in   
                [x in collect(distinct {class:class, sdc:sdc, r:r}) WHERE NOT x['r'] IS NULL] | //filtering for classes with existing Column MAPS_TO_CLASS relationship
                [
                    //y['class'].label + '.' + y['class'].short_label,    //to be used as key of the dict
                    y['class'].short_label,    //to be used as key of the dict
                    y['sdc']._columnname_                               // to be used as value of the dict
                ]
            ]
        ) as rename_dct,
        apoc.map.fromPairs(collect([sdc._columnname_, sdc.Order])) as order_dct,
        sdt['SortOrder'] as sorting
        """

        params = {'standard': standard, 'table': table}
        if self.debug:
            logging.debug(f"""
                               query: {q}
                               parameters: {params}
                           """)
        res = self.query(q, params)
        return res

    def neo_get_mapped_classes(self):
        q = """
        MATCH (sdf:`Source Data Folder`)-[:HAS_TABLE]->(sdt:`Source Data Table`),
        (sdt)-[:HAS_COLUMN]->(sdc:`Source Data Column`),
        (sdc)-[:MAPS_TO_PROPERTY]->(property:Property),
        (class:Class)-[:HAS_PROPERTY]->(property:Property)               
        RETURN DISTINCT class.label as Class
        """
        params = {}
        if self.debug:
            logging.debug(f"""
                               query: {q}
                               parameters: {params}
                           """)
        res = self.query(q, params)
        if res:
            return [r['Class'] for r in res]
        else:
            return []

    def filter_classes_from_rels(self, rels, classes: list):
        clean_rels = []
        excluded_rels = []
        for rels_dict in rels:
            intersect = False
            for class_ in classes:
                if class_ in rels_dict.values():
                    intersect = True

            if not intersect:
                clean_rels.append(rels_dict)
            else:
                excluded_rels.append(rels_dict)

        if excluded_rels:
            print(f'The following rels were excluded: {excluded_rels}')

        return clean_rels

    def neo_validate_classes_to_extract(self, classes: list) -> ([], []):
        """
        :param classes: list of classes to validate (nodes with labels Class and property label must exist)
        :return: [], [] - list of valid classes and list if invalid classes
        Identifies non-valid classes with count==0 or no count property (at least 1 instance got extracted during reshaping)
        """
        q = """
        MATCH (n:Class) 
        WHERE n.count = 0 OR NOT EXISTS (n.count)
        RETURN DISTINCT n.label as class
        """
        pre_non_valid = [res['class'] for res in self.query(q, {'classes': classes})]
        non_valid = []
        for class_ in pre_non_valid:
            if class_ in classes:
                non_valid.append(class_)

        valid = [class_ for class_ in classes if class_ not in non_valid]
        return valid, non_valid

    def neo_validate_access(self, classes: list, user_role=None) -> ([], []):
        """
        :param classes: list of classes to validate (nodes with labels Class and property label must exist)
        :param user_role: when None no access restrictions are accounted for, otherwise the classes with
        (:`User Role`{name:$user_role})-[:RESTRICTED_ACCESS]->(:Class) would be considered invalid
        :return:
        Identifies non-valid classes to which the user_role(if provided does not have access to)
        """
        no_access = []
        if user_role:
            q = """
                MATCH (role:`User Role`{name:$user_role})
                OPTIONAL MATCH (role)-[:ACCESS_RESTRICTED]->(c:Class)
                RETURN DISTINCT role.name as role, c.label as class
                """
            pre_no_access = [res['class'] for res in self.query(q, {'user_role': user_role})]
            if pre_no_access == []:
                raise Exception(f"User Role {user_role} does not exist")
            for class_ in pre_no_access:
                if class_ in classes:
                    no_access.append(class_)
        has_access = [class_ for class_ in classes if class_ not in no_access]
        return has_access, no_access

    @staticmethod
    def check_dataframes_equal(df1, df2):
        compare = datacompy.Compare(df1.reset_index(), df2.reset_index(), join_columns=['index'])
        return compare.report()



import pandas as pd
import pytest
import datacompy

from cdisc_data_providers.sdtm_data_provider import SDTMDataProvider
@pytest.fixture(scope="module")
def dp():
    pass
    # dp = data_provider.DataProvider()
    #dp.clean_slate()   # At times it causes trouble!
    # yield dp

def test_check_dataframes_equal():
    df1 = pd.DataFrame({1: ['a', 'b', 'c'], 2: ['c', 'd', 'e']})
    df2 = pd.DataFrame({1: ['a', 'b', 'c'], 2: ['c', 'd', 'e']})
    expected_result = datacompy.Compare(df1.reset_index(), df2.reset_index(), join_columns=['index'])
    result = SDTMDataProvider.check_dataframes_equal(df1, df2)

    assert expected_result.report() == result



import pytest
from cdisc_data_providers import sdtm_data_provider
import pandas as pd
import json
import os

filepath = os.path.dirname(__file__)


# Provide a DataProvider object (which contains a database connection)
@pytest.fixture(scope="module")
def dp():
    dp = sdtm_data_provider.SDTMDataProvider(debug=True, check_for_refarctored=True)
    yield dp


def test_get_data_sdtm_dm1(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)
    dp.verbose = True

    standard = 'test_standard'
    table = 'DM'
    res = dp.get_data_sdtm(standard=standard, domain=table)
    expected = pd.DataFrame({'TC4': ['test_data_2', 'test_data_1'], 'TC5': [None, None], 'TC3': ['test_study_1', 'test_study_2']})

    pd.testing.assert_frame_equal(res, expected)


def test_get_data_sdtm_dm2(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)
    dp.verbose = True

    standard = 'test_standard'
    table = 'AE'
    res = dp.get_data_sdtm(standard=standard, domain=table)
    expected = pd.DataFrame({'TC1': ['test_data_1', 'test_data_2'], 'TC2': ['test_study_1', 'test_study_2']})

    pd.testing.assert_frame_equal(res, expected)


def test_get_data_restricted_access(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm_restricted.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)
    dp.verbose = True

    standard = 'test_standard'
    table = 'AE'
    res = dp.get_data_sdtm(standard=standard, domain=table, user_role='test_role')
    expected = pd.DataFrame({'TC1': [None, None], 'TC2': ['test_study_3', 'test_study_4']})

    pd.testing.assert_frame_equal(res, expected)


def test_neo_get_meta_dm1(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)

    # AE domain
    standard = 'test_standard'
    table = 'AE'
    res = dp.neo_get_meta(standard=standard, table=table)[0]

    rels = res.get('rels')
    expected_rels = [
        {'short_label': 'STUDYID', 'from': 'Adverse Events', 'to': 'Study', 'type': 'test_rel_1'},
        {'short_label': 'tc2', 'from': 'Adverse Events', 'to': 'test_class_2', 'type': 'test_rel_2'}
    ]
    assert [i for i in expected_rels if i not in rels] == []
    assert len(rels) == len(expected_rels)

    classes = res.get('classes')
    expected_classes = ['Study', 'Adverse Events', 'test_class_2']
    assert sorted(classes) == sorted(expected_classes)

    assert res.get('req_classes') == ['Study']
    assert sorted(res.get('order_dct')) == sorted({'TC1': 1, 'TC2': None})
    assert res.get('sorting') == ['TC1', 'TC2']


def test_neo_get_meta_dm2(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)

    # AE domain
    standard = 'test_standard'
    table = 'DM'
    res = dp.neo_get_meta(standard=standard, table=table)[0]

    rels = res.get('rels')
    expected_rels = [
        {'short_label': 'STUDYID', 'from': 'Demographics', 'to': 'Study', 'type': 'test_rel_3'},
        {'short_label': 'tc7', 'from': 'Demographics', 'to': 'test_class_7', 'type': 'test_rel_5'},
        {'short_label': 'tc6', 'from': 'Demographics', 'to': 'test_class_6', 'type': 'test_rel_4'}
    ]
    assert [i for i in expected_rels if i not in rels] == []
    assert len(rels) == len(expected_rels)

    classes = res.get('classes')
    expected_classes = ['Study', 'test_class_6', 'test_class_7', 'Demographics']
    assert sorted(classes) == sorted(expected_classes)

    assert sorted(res.get('req_classes')) == sorted(['test_class_6'])
    assert sorted(res.get('order_dct')) == sorted({'TC5': 3, 'TC4': 2, 'TC3': None})
    assert res.get('sorting') == ['TC3', 'TC4', 'TC5']


def test_neo_get_mapped_classes(dp):
    dp.clean_slate()
    with open(os.path.join(filepath, 'data', 'test_data_sdtm_mapped_classes.json')) as jsonfile:
        dct = json.load(jsonfile)
    dp.load_arrows_dict(dct)
    res = dp.neo_get_mapped_classes()
    expected = ['test_label_1', 'test_label_2']
    difference = set(res) ^ set(expected)
    assert not difference


def test_neo_validate_classes_to_extract(dp):
    dp.clean_slate()

    valid_labels = ['test_label_1', 'test_label_2']
    invalid_labels = ['test_label_3', 'test_label_4']
    q = """
    MERGE (:Class {label: 'test_label_1', count: 10})
    MERGE (:Class {label: 'test_label_2', count: 1})
    MERGE (:Class {label: 'test_label_3', count: 0})
    MERGE (:Class {label: 'test_label_4'})
    """
    dp.query(q)

    # All valid
    assert dp.neo_validate_classes_to_extract(classes=valid_labels) == (valid_labels, [])
    # Some valid
    assert dp.neo_validate_classes_to_extract(classes=valid_labels+invalid_labels) == (valid_labels, invalid_labels)
    # None valid
    assert dp.neo_validate_classes_to_extract(classes=invalid_labels) == ([], invalid_labels)


def test_neo_validate_access(dp):
    dp.clean_slate()
    classes = ["End Date/Time of Observation", "Date/Time of Reference Time Point", "Date/Time of Collection",
               "Subject", "Race"]
    dp.query(
        """
        UNWIND $classes as class
        MERGE (c:Class{label:class})
        """,
        {'classes': classes}
    )
    dp.query("""
    WITH [['my_login', 'Study Lead'], ['external', 'External Researcher']] as coll
    UNWIND coll as pair
    WITH pair[0] as user_id, pair[1] as role_name
    MERGE (user:User{id:user_id})-[r1:HAS_ROLE]->(role:`User Role`{name:role_name})
    WITH *
    WHERE role.name = 'External Researcher'
    MATCH (class:Class)
    WHERE class.label in $rclasses
    MERGE (role)-[:ACCESS_RESTRICTED]->(class)
    """,
    {'rclasses': classes[:3]}
    )

    # External Researcher
    res_has_access, res_no_access = dp.neo_validate_access(classes=classes, user_role="External Researcher")
    print(res_has_access, res_no_access)
    assert set(res_has_access) == {'Subject', 'Race'}
    assert set(res_no_access) == {'Date/Time of Reference Time Point', 'Date/Time of Collection', 'End Date/Time of Observation'}

    # Study Lead
    res_has_access, res_no_access = dp.neo_validate_access(classes=classes, user_role="Study Lead")
    assert set(res_has_access) == {'Subject', 'Race', 'Date/Time of Reference Time Point', 'Date/Time of Collection',
                                  'End Date/Time of Observation'}
    assert len(res_no_access) == 0

    # Not specified
    res_has_access, res_no_access = dp.neo_validate_access(classes=classes)
    assert set(res_has_access) == {'Subject', 'Race', 'Date/Time of Reference Time Point', 'Date/Time of Collection',
                                  'End Date/Time of Observation'}
    assert len(res_no_access) == 0

    # Role does not exist
    with pytest.raises(Exception):
        dp.neo_validate_access(classes=classes, user_role="Directors")



import os
from model_appliers.model_applier import ModelApplier
import json
from data_loaders import file_data_loader
import re


class CdiscStandardLoader(ModelApplier):

    def __init__(self, standards_folder: str = None, sdtm_file: str = None, sdtmig_file: str = None, terminology_file: str = None,
                 *args, **kwargs):
        """
        :param standards_folder: Directory where standard files are stored
        :param sdtm_file: Name of file containing SDTM Model metadata
        :param sdtmig_file: Name of file containing SDTMIG metadata
        :param terminology_file: Name of file containing SDTM Terminology
        """
        super().__init__(rdf=True, *args, **kwargs)
        assert os.path.exists(standards_folder)
        assert os.path.exists(os.path.join(standards_folder, sdtm_file))
        assert os.path.exists(os.path.join(standards_folder, sdtmig_file))
        assert os.path.exists(os.path.join(standards_folder, terminology_file))
        self.standards_folder = standards_folder
        self.sdtm_file = sdtm_file
        self.sdtmig_file = sdtmig_file
        self.terminology_file = terminology_file

    def load_standard(self, extract_terms: bool = True, extract_vld: bool = True):
        print("Loading content")
        print("Standards folder:", self.standards_folder)
        print("Standards model:", self.sdtm_file)
        print("Standards Implementation Guide:", self.sdtmig_file)
        print("Terminology file:", self.terminology_file)

        fdl = file_data_loader.FileDataLoader()
        df = fdl.load_file(self.standards_folder, self.sdtm_file)
        q = f"""
        MATCH (n:`Source Data Row`)
        REMOVE n:`Source Data Row`
        SET    n:`{self.sdtm_file}`
        """
        self.query(q)

        df = fdl.load_file(self.standards_folder, self.sdtmig_file)
        q = f"""
        MATCH (n:`Source Data Row`)
        REMOVE n:`Source Data Row`
        SET    n:`{self.sdtmig_file}`
        """
        self.query(q)

        df = fdl.load_file(self.standards_folder, self.terminology_file)
        q = f"""
        MATCH (n:`Source Data Row`)
        REMOVE n:`Source Data Row`
        SET    n:`{self.terminology_file}`
        """
        self.query(q)

        with open(os.path.join(self.standards_folder, "sdtmig3_2_domain_sort_order.json"), 'r') as json_file:
            json_data = json.load(json_file)
        self.DOMAIN_SORT_ORDER = json_data["domain_sort_order"]
        with open(os.path.join(self.standards_folder, "sdtmig3_3_domain_labels.json"), 'r') as json_file:
            json_data = json.load(json_file)
        self.DOMAIN_LABELS = json_data["domain_labels"]
        self.reshape_model()
        self.reshape_sdtmig()
        self.reshape_terminology()
        self.link_cdisc()
        self.load_link_sdtm_ttl(local=False)

    def reshape_model(self):
        # Create nodes for General Observation Classes
        q = f"""
        MATCH (n:`{self.sdtm_file}`)
        WITH collect(distinct(n.Class)) as gocs
        UNWIND gocs as goc
        MERGE (d:ObservationClass {{Class: toUpper(goc), label:goc}})
        RETURN count(d)
        """
        self.query(q)

        # Change label on general observation class variables
        q = f"""
        MATCH (n:`{self.sdtm_file}`)
        WHERE n.`Dataset Name` = ""
        REMOVE n:`{self.sdtm_file}`
        SET    n:GOC
        """
        self.query(q)

        # Change label on special purpose class variables
        q = f"""
        MATCH (n:`{self.sdtm_file}`)
        WHERE n.`Dataset Name` <> ""
        REMOVE n:`{self.sdtm_file}`
        SET    n:Special_Purpose_Variable
        """
        self.query(q)

        # Link General Observation Class to variables GOC-[CLASS_SPECIFIC_VARIABLE_GROUPING]-(VariableGrouping)
        # First create variable grouping node
        q = """
        MATCH (oc:ObservationClass)
        MERGE (vg:VariableGrouping {contextLabel: oc.label + " Observation Class Variables"})
        SET vg.contextLabel = oc.label + " Observation Class Variables"
        MERGE (oc)-[:CLASS_SPECIFIC_VARIABLE_GROUPING]->(vg)
        RETURN count(vg)
        """
        self.query(q)

        # Link General Observation Class to variable grouping node
        q = """
        MATCH (oc:ObservationClass)-[:CLASS_SPECIFIC_VARIABLE_GROUPING]->(vg:VariableGrouping)
        MATCH (v:GOC)
        WHERE v.Class = oc.label
        MERGE (v)-[:context]->(vg)
        return count(v)
        """
        self.query(q)

    def reshape_sdtmig(self):
        # Change label on domain variables
        # Adaptions for load_link_sdtm_ttl
        # - Add property Variable (Variable Name) (lls)
        # Adaption for generate model
        # - Add property Label (Variable Label)
        q = f"""
        MATCH (n:`{self.sdtmig_file}`)
        REMOVE n:`{self.sdtmig_file}`
        SET    n:Variable
        SET    n.Variable = n.`Variable Name` 
        SET    n.Label = n.`Variable Label` 
        """
        self.query(q)

        # Add domains/dataset
        # Adaptions for load_link_sdtm_ttl
        # - Add label Dataset
        q = f"""
        MATCH (n:Variable)
        WITH collect(distinct(n.`Dataset Name`)) as domains
        UNWIND domains as domain
        MERGE (d:Domain:Dataset {{Domain: domain, Dataset: domain}})
        """
        self.query(q)

        # Relate variables to domain and set class of domain
        q = f"""
        MATCH (d:Domain)
        MATCH (v:Variable)
        WHERE d.Domain = v.`Dataset Name`
        MERGE (d)-[:HAS_VARIABLE]->(v)
        SET d.Class = toUpper(v.Class)
        RETURN count(v)
        """
        self.query(q)

        # Remove empty value list
        q = """
        MATCH (n:Variable)
        WHERE n.`Value List` = ""
        REMOVE n.`Value List`
        RETURN count(n)
        """
        self.query(q)

        # Add property for codelist to variable
        # Adapt for generate_excel_based_model
        # - Add property Label (Variable Label)
        q = """
        MATCH (v:Variable)
        WHERE v.`CDISC CT Codelist Code(s)` <> ""
        SET v.`Codelist Code` = v.`CDISC CT Codelist Code(s)`
        SET v.Label = v.`Variable Label`
        RETURN count(v)
        """
        self.query(q)

        # Add domain labels
        # Adapt for generate_excel_based_model
        # - Add property Description (Label)
        for [domain, label] in self.DOMAIN_LABELS.items():
            q = f"""
            MATCH (d:Domain)
            WHERE d.Domain = $domain
            SET d.label = $label
            SET d.Description = $label
            RETURN d.label
            """
            if (self.debug == True):
                print(domain, label)
                print(q)
            self.query(q, {'domain': domain, 'label': label})

        # Link domains to General Observation Classes
        q = f"""
        MATCH (d:Domain)
        MATCH (g:ObservationClass)
        WHERE g.Class = d.Class
        MERGE (g)-[:HAS_DATASET]->(d)
        RETURN count(d)
        """
        self.query(q)

    def reshape_terminology(self):
        # Change label on codelists
        q = f"""
        MATCH (n:`{self.terminology_file}`)
        WHERE n.`Codelist Code` = ""
        REMOVE n:`{self.terminology_file}`
        SET    n:Codelist
        """
        self.query(q)

        # Change label on codelist items
        # Adapt for generate_excel_based_model
        # - Add property Term (CDISC Submission Value)
        # - Add property `NCI Codelist Code` (Codelist Code)
        # - Add property `NCI Term Code` (Code)
        q = f"""
        MATCH (n:`{self.terminology_file}`)
        WHERE n.`Codelist Code` <> ""
        REMOVE n:`{self.terminology_file}`
        SET    n:Term
        SET    n.Term = n.`CDISC Submission Value`
        SET    n.`NCI Codelist Code` = n.`Codelist Code`
        SET    n.`NCI Term Code` = n.Code
        """
        self.query(q)

        # Link codelist item to codelist
        q = f"""
        MATCH (c:Codelist)
        MATCH (t:Term)
        WHERE t.`Codelist Code` = c.Code
        MERGE (c)-[:HAS_TERM]->(t)
        RETURN count(t)
        """
        self.query(q)

    def link_cdisc(self, extract_terms: bool = True, extract_vld: bool = True):
        print("Linking CDISC content")
        for [domain, sort_order] in self.DOMAIN_SORT_ORDER.items():
            # hardcode adding --SEQ into the 'Sort Order'
            if (self.debug == True):
                print(domain, sort_order)
            q = f"""
            MATCH (n:Domain)-[:HAS_VARIABLE]->(v:Variable)
            WHERE v.`Variable Name` ends with "SEQ" AND n.Domain = "{domain}"
            WITH  n, v.`Variable Name` as seq
            SET n.`Sort Order` = CASE WHEN seq in apoc.text.split("{sort_order}", ",") THEN "{sort_order}" ELSE "{sort_order}" + "," + seq END
            RETURN n["Sort Order"]
            """
            self.query(q)

            # Variables need to have a property Order = sort order
            variables = sort_order.split(",")
            # Add --SEQ if not mentioned.
            add_seq = [var for var in variables if re.findall('SEQ', var.upper())]
            # N.B! Only to applicable to domains that can have --SEQ variable. Not trial design DM and SV.
            if not add_seq and not domain in ['DM', 'SV', 'TA', 'TE', 'TV', 'TS', 'TI']:
                variables.append(domain+'SEQ')

            for i, var in enumerate(variables, start=1):
                q = f"""
                MATCH (v:Variable)
                WHERE v.`Dataset Name` = '{domain}' AND v.`Variable Name` = '{var}'
                SET v.Order = {i}
                RETURN v.Order
                """
                self.query(q)

        # Add relationship to Codelist for variable
        # CREATE (v:Variable)-[:HAS_CONTROLLED_TERM]->(t:Term) Info on exact terms does not exist
        q = """
        MATCH (v:Variable)
        MATCH (c:Codelist)
        WHERE v.`Codelist Code` = c.Code
        MERGE (v)-[:HAS_CODELIST]->(c)
        WITH *
        MATCH (c)-[:HAS_TERM]->(t:Term)
        MERGE (v)-[:HAS_CONTROLLED_TERM]->(t)
        RETURN COUNT(v)
        """
        self.query(q)

        # Add relationship to terms for variable
        # TODO: This might not be needed. Adds variable.`Value List` relationship to terms
        # 1. Add for when VALUE LIST is a semi-colon separated list
        # NB: Extra space inserted before semi-colon, so splitting on "; "
        q = """
        MATCH (v:Variable)
        WHERE v.`Value List` contains ";"
        WITH v, apoc.text.split(v.`Value List`, "; ") as  terms
        UNWIND terms as trm
        MATCH (t:Term) WHERE t.`CDISC Submission Value` = trm
        MERGE (v)-[:HAS_CONTROLLED_TERM]->(t)
        RETURN count(v), count(t)
        """
        self.query(q)
        # 2. Add for when VALUE LIST is just a singe term
        # NB: The only values in SDTMIG 3.2 are domain codes (e.g. DM, AE etc.) and they do not exist in CT downloaded from CDISC Library
        #     So this statement has no effect at all at the moment.
        q = """
        MATCH (v:Variable)
        WHERE NOT v.`Value List` contains ";"
        MATCH (t:Term)
        WHERE t.`CDISC Submission Value` = v.`Value List`
        MERGE (v)-[:HAS_CONTROLLED_TERM]->(t)
        RETURN v, t
        """
        self.query(q)

        if extract_terms:
            # SET t.`Codelist Code` = t.Code # Codelist Code is already assigned earlier
            q = """
            MATCH (t:Term)       
            SET t.`Term Code` = t.Code
            """
            self.query(q)

            # TODO: No Term is linked to Variable (HAS_CONTROLLED_TERM), so this will not do anything
            # # #merging duplicate Terms together
            # q = """
            # MATCH (t:Term)
            # WITH t.`Codelist Code` as cl, t.`Term Code` as trm, collect(t) as coll
            # CALL apoc.refactor.mergeNodes(coll) YIELD node
            # REMOVE node.ID
            # WITH node
            # MATCH (node)<-[r:HAS_CONTROLLED_TERM]-(c)
            # WITH c, node, collect(r) as coll
            # WHERE size(coll)>1
            # WITH coll[1..] as coll
            # UNWIND coll as r
            # DELETE r
            # """
            # self.query(q)

        # TODO:
        #  1. As above, HAS_CONTROLLED_TERM relationship does not exist in CDISC metadata
        #  2. CDISC CT does not have a property Order for terms
        # # link related Terms sequentially according to their Order property
        # q = """
        # MATCH (v:Variable)-[:HAS_CONTROLLED_TERM]->(t:Term)
        # WITH v,t ORDER BY v.label, t.Order, t.Term ASC
        # WITH v, COLLECT(t) AS terms
        # FOREACH (n IN RANGE(0, SIZE(terms)-2) |
        #     FOREACH (prev IN [terms[n]] |
        #         FOREACH (next IN [terms[n+1]] |
        #             MERGE (prev)-[:NEXT]->(next))))
        # """
        # self.query(q)

        print("CDISC Data Link Complete")

    def load_link_sdtm_ttl(self, local=True):
        self.rdf_config()
        if local:
            with open(os.path.join(self.standards_folder, 'sdtm-1-3.ttl')) as f:
                rdf = f.read()
                print(
                    self.rdf_import_subgraph_inline(rdf, "Turtle")
                )
        else:
            print(
                self.rdf_import_fetch(
                    "https://raw.githubusercontent.com/phuse-org/rdf.cdisc.org/master/std/sdtm-1-3.ttl",
                    "Turtle"
                )
            )

        self.create_index(label="DataElement", key="dataElementName")

        # linking ObservationClass to VariableGrouping (only class specific)
        q = """
        MATCH (oc:ObservationClass), (vg:VariableGrouping)
        WHERE vg.contextLabel = 
            CASE oc.Class 
                WHEN "FINDINGS" THEN "Findings Observation Class Variables"
                WHEN "EVENTS" THEN "Event Observation Class Variables"
                WHEN "INTERVENTIONS" THEN "Interventions Observation Class Variables"
            END
        MERGE (oc)-[:CLASS_SPECIFIC_VARIABLE_GROUPING]->(vg)
        """
        self.query(q)

        # adding properties dataElementName
        q = """ 
        MATCH (v:Variable), (de:DataElement)
        WHERE v.Variable =~ apoc.text.replace(de.dataElementName, '-', '.')
           AND NOT (v.Variable STARTS WITH 'RF' and v.Dataset = 'DM')
        WITH DISTINCT v, de.dataElementName as dataElementName, de.dataElementLabel as dataElementLabel
        SET v.dataElementName = dataElementName, v.dataElementLabel = dataElementLabel 
        """
        self.query(q)

        # linking Variable to DataElements (on dataElementName there might be >1 DataElement per Variable -
        # need to filter on VariableGrouping~ObservationClass)
        # (1) WHERE size(coll)=1
        q = """
        MATCH (v:Variable)<-[r:HAS_VARIABLE]-(ds:Dataset)    
        OPTIONAL MATCH (da:DataElement)
        WHERE da.dataElementName = v.dataElementName        
        WITH v, collect(da) as coll
        WHERE size(coll)=1
        UNWIND coll as da
        MERGE (v)-[:IS_DATA_ELEMENT]->(da)
        """
        self.query(q)

        # 2 else
        q = """
        MATCH (v:Variable)<-[r:HAS_VARIABLE]-(ds:Dataset)<-[:HAS_DATASET]-(oc:ObservationClass)
        , (da:DataElement)-[:context]->(vg:VariableGrouping)<-[:CLASS_SPECIFIC_VARIABLE_GROUPING]-(oc)        
        WHERE NOT EXISTS ((v)-[:IS_DATA_ELEMENT]->()) AND da.dataElementName = v.dataElementName        
        MERGE (v)-[:IS_DATA_ELEMENT]->(da)               
        """
        self.query(q)

        # saveing counts of data elements with the same name
        q = """
        MATCH (da:DataElement)
        OPTIONAL MATCH (da:DataElement)-[:context]->(vg:VariableGrouping)
        WITH *, 
        {
          `Interventions Observation Class Variables`: "Interventions",
          `General Observation Class Timing Variables`: "GO Timing",
          `Findings Observation Class Variables`: "Findings",
          `General Observation Class Identifier Variables`: "GO Identifier",
          `Event Observation Class Variables`: "Events",
          `Findings About Events or Interventions Variables`: "FA"
        } as vg_mapping 
        SET da.vg = vg.contextLabel                
        SET da.vg_short = vg_mapping[vg.contextLabel]
        WITH DISTINCT da.dataElementName as name, collect(da) as coll
        WITH *, size(coll) as sz
        UNWIND coll as da
        SET da.n_with_same_name = sz 
        """
        self.query(q)

        # saveing counts of variables with the same label
        q = """
        MATCH (v:Variable)
        WITH v.Label as lbl, collect(v) as coll
        WITH *, size(coll) as sz
        UNWIND coll as v
        SET v.n_with_same_label = sz
        """
        self.query(q)

        # saveing counts of variables with the same name
        q = """
        MATCH (v:Variable)
        WITH v.Variable as name, collect(v) as coll
        WITH *, size(coll) as sz
        UNWIND coll as v
        SET v.n_with_same_name = sz
        """
        self.query(q)

        print("SDTM TTL Loaded and Linked")

    def propagate_relationships(self, on_children=True, on_parents=True):  # not used kept for code reference
        la = ('' if (on_children and on_parents) or not on_children else '<')
        ra = ('' if (on_children and on_parents) or not on_parents else '>')
        q = f"""
        //propagate_relationships_of_parents_on_children
        MATCH (c:Class)
        OPTIONAL MATCH path = (c){la}-[:SUBCLASS_OF*1..50]-{ra}(parent)<-[r1:FROM]-(r:Relationship)-[r2:TO]->(fromto)
        WITH c, collect(path) as coll
        OPTIONAL MATCH path = (c){la}-[:SUBCLASS_OF*1..50]-{ra}(parent)<-[r1:TO]-(r:Relationship)-[r2:FROM]->(fromto)
        WITH c, coll + collect(path) as coll
        UNWIND coll as path
        WITH 
            c, 
            nodes(path)[-1] as fromto, 
            nodes(path)[-2] as r,
            relationships(path)[-1] as r2, 
            relationships(path)[-2] as r1
        WITH *, apoc.text.join([k in [x in keys(r) WHERE x <> 'uri'] | '`' + k + '`: "' + r[k] + '"'], ", ") as r_params  
        WITH *, CASE WHEN r_params = '' THEN '' ELSE '{' + r_params + '}' END as r_params
        WITH *, 
         '
                WITH $c as c
                MATCH (fromto) WHERE id(fromto) = $id_fromto
                MERGE (c)<-[:`'+type(r1)+'`]-(:Relationship' + r_params +')-[:`'+type(r2)+'`]->(fromto) 
         ' as q 
        WITH c, fromto, q
        call apoc.do.when(
            NOT EXISTS ( (c)--(:Relationship)--(fromto) ), 
            q,
            '',    
            {{c:c, id_fromto:id(fromto)}}
        ) YIELD value
        RETURN value 
        """
        self.query(q)
